/*
Copyright (C) 2020 Jacquelin POTIER <jacquelin.potier@free.fr>
Dynamic aspect ratio code Copyright (C) 2020 Jacquelin POTIER <jacquelin.potier@free.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//-----------------------------------------------------------------------------
// Object: main
//-----------------------------------------------------------------------------

#pragma once

#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0501 // for xp os
#endif

#include "../Tools/gui/NotifyIcon/NotifyIcon.h"
#include "../Tools/gui/Menu/PopUpMenu.h"
#include "../Tools/Process/SingleInstance/CSingleInstance.h"

#include <windows.h>
#include "resources/resource.h"


// Enable Visual Style
#if defined _AMD64
#pragma comment(linker,"/manifestdependency:\"type='win32' name='Microsoft.Windows.Common-Controls' version='6.0.0.0' processorArchitecture='amd64' publicKeyToken='6595b64144ccf1df' language='*'\"")
#elif defined _WIN64
#pragma comment(linker,"/manifestdependency:\"type='win32' name='Microsoft.Windows.Common-Controls' version='6.0.0.0' processorArchitecture='ia64' publicKeyToken='6595b64144ccf1df' language='*'\"")
#else
#pragma comment(linker,"/manifestdependency:\"type='win32' name='Microsoft.Windows.Common-Controls' version='6.0.0.0' processorArchitecture='x86' publicKeyToken='6595b64144ccf1df' language='*'\"")
#endif

#define START_DISABLING TEXT("Start Disabling Screensaver")
#define STOP_DISABLING  TEXT("Stop Disabling Screensaver")
#define DisableScreenSaverWebSite TEXT("http://jacquelin.potier.free.fr/disablescreensaver")
#define CDisableScreenSaver_CLASS_NAME  TEXT("CDisableScreenSaverClassName")

class CDisableScreenSaver:public CNotifyIconEventsBase
{
protected:
    HINSTANCE hInstance;
    HWND hWndDialog;
    CNotifyIcon* pNotifyIcon;
    CPopUpMenu* pPopUpMenu;
    HICON hIconStartDisabling;
    HICON hIconStopDisabling;

    UINT MenuIdStartStop;
    UINT MenuIdWebSite;
    UINT MenuIdExit;

    BOOL bScreenSaverDisabled;

    BOOL NotifyIconCallBack(LPPOINT pCursorPos);
    virtual BOOL OnNotifyIconClick(CNotifyIcon* pSender,LPPOINT pCursorPos){UNREFERENCED_PARAMETER(pSender);return this->NotifyIconCallBack(pCursorPos);}
    virtual BOOL OnNotifyIconRightClick(CNotifyIcon* pSender,LPPOINT pCursorPos){UNREFERENCED_PARAMETER(pSender);return this->NotifyIconCallBack(pCursorPos);}
    virtual BOOL OnNotifyIconDoubleClick(CNotifyIcon* pSender,LPPOINT pCursorPos){UNREFERENCED_PARAMETER(pSender);return this->NotifyIconCallBack(pCursorPos);}

    void OnClose();
    // void OnInit();
    static CDisableScreenSaver* GetAssociatedObject(HWND hWndDialog);
    BOOL WaitApplicationEnd();
    BOOL Exit();

    static LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

    void OnStartStop();
public:
    CDisableScreenSaver(HINSTANCE hInstance);
    ~CDisableScreenSaver();
    
    BOOL Run();
};

int WINAPI WinMain (HINSTANCE hInstance,
                    HINSTANCE hPrevInstance,
                    LPSTR lpCmdLine,
                    int nCmdShow
                    )
{
    UNREFERENCED_PARAMETER(hPrevInstance);
    UNREFERENCED_PARAMETER(lpCmdLine);
    UNREFERENCED_PARAMETER(nCmdShow);

    InitCommonControls();

    CSingleInstance SingleInstance(TEXT("DisableScreenSaverJP"));
    if (SingleInstance.IsAnotherInstanceRunning())
    {
        ::MessageBox(NULL,
                    TEXT("Another instance is already running. Look in your traybar (near the time display) to change the settings or quit the existing application"),
                    TEXT("DisableScreenSaver"),
                    MB_ICONINFORMATION |MB_OK
                    );
        return 0;
    }

    CDisableScreenSaver DisableScreenSaver(hInstance);
    if (!DisableScreenSaver.Run())
        return -1;

    return 0;
}

void CDisableScreenSaver::OnStartStop()
{
    // SetThreadExecutionState flags
    //   ES_SYSTEM_REQUIRED Informs the system that the thread is performing some operation that is not normally detected as activity by the system. 
    //   ES_DISPLAY_REQUIRED Informs the system that the thread is performing some operation that is not normally detected as display activity by the system. 
    //   ES_USER_PRESENT Informs the system that a user is present. If a user is present, the system will use the power policy settings set by the user. Otherwise, the system does not wake the display device and will return to the sleeping state as soon as possible. 
    //   ES_AWAYMODE_REQUIRED Enables away mode. This value must be specified with ES_CONTINUOUS.
    //          Away mode should be used only by media-recording and media-distribution applications that must perform critical background processing on desktop computers while the computer appears to be sleeping.
    //          value should be used only when absolutely necessary by media applications that require the system to perform background tasks such as recording television content or streaming media to other devices while the system appears to be sleeping. Applications that do not require critical background processing or that run on portable computers should not enable away mode because it prevents the system from conserving power by entering true sleep.
    //   ES_CONTINUOUS Informs the system that the state being set should remain in effect until the next call that uses ES_CONTINUOUS and one of the other state flags is cleared.
    //#define ES_SYSTEM_REQUIRED   ((DWORD)0x00000001)
    //#define ES_DISPLAY_REQUIRED  ((DWORD)0x00000002)
    //#define ES_USER_PRESENT      ((DWORD)0x00000004)
    //#define ES_AWAYMODE_REQUIRED ((DWORD)0x00000040)
    //#define ES_CONTINUOUS        ((DWORD)0x80000000)
    // windows media player usage of SetThreadExecutionState (thx to WinApiOverride)
    //  video start
    //    -SetThreadExecutionState(0x0000000080000003)		wmp.dll + 0x3A0481	C:\WINDOWS\SYSTEM32\wmp.dll	
    //  video stop
    //    -SetThreadExecutionState(0x0000000080000000)		wmp.dll + 0x3A04BE	C:\WINDOWS\SYSTEM32\wmp.dll	
    EXECUTION_STATE OldState;
    EXECUTION_STATE NewState;

    HICON NewMenuIcon;
    TCHAR* NewMenuText;

    if (this->bScreenSaverDisabled)
    {
        // stop disabling
        NewState = ES_CONTINUOUS;

        NewMenuIcon = this->hIconStartDisabling;
        NewMenuText = START_DISABLING;
    }
    else
    {
        // start disabling
        NewState = ES_CONTINUOUS | ES_DISPLAY_REQUIRED | ES_SYSTEM_REQUIRED;

        NewMenuIcon = this->hIconStopDisabling;
        NewMenuText = STOP_DISABLING;
    }

    this->pPopUpMenu->SetIcon(this->MenuIdStartStop,NewMenuIcon);
    this->pPopUpMenu->SetText(this->MenuIdStartStop,NewMenuText);
    OldState = SetThreadExecutionState(NewState);

    // invert state
    this->bScreenSaverDisabled = !this->bScreenSaverDisabled;
}

BOOL CDisableScreenSaver::NotifyIconCallBack(LPPOINT pCursorPos)
{
    UINT ClickedMenuId;
    ClickedMenuId = this->pPopUpMenu->Show(pCursorPos->x,pCursorPos->y,this->pNotifyIcon->GetAssocatedNotifyIconHwnd(),FALSE,TRUE);
    if (ClickedMenuId)
    {
        if (ClickedMenuId == this->MenuIdStartStop)
        {
            this->OnStartStop();
        }
        else if (ClickedMenuId == this->MenuIdWebSite)
        {
            if (((int)ShellExecute(NULL,_T("open"),DisableScreenSaverWebSite,NULL,NULL,SW_SHOWNORMAL))<33)
                return FALSE;
        }
        else if (ClickedMenuId == this->MenuIdExit)
        {
            this->Exit();
        }
    }

    // return TRUE to restore focus to notify icon, FALSE to keep focus
    return FALSE;
}


CDisableScreenSaver::CDisableScreenSaver(HINSTANCE hInstance)
{
    this->hInstance = hInstance;
    this->hWndDialog = NULL;
    this->pPopUpMenu = NULL;

    this->MenuIdStartStop = (UINT)-1;
    this->MenuIdWebSite = (UINT)-1;
    this->MenuIdExit = (UINT)-1;

    this->bScreenSaverDisabled = FALSE;
}
CDisableScreenSaver::~CDisableScreenSaver()
{

}

BOOL CDisableScreenSaver::Run()
{
    // create notify icon
    this->hIconStartDisabling = (HICON)::LoadImage(this->hInstance,MAKEINTRESOURCE(IDI_ICON_START_DISABLING_SCREEN_SAVER),IMAGE_ICON,16,16,LR_DEFAULTCOLOR|LR_SHARED);
    this->hIconStopDisabling = (HICON)::LoadImage(this->hInstance,MAKEINTRESOURCE(IDI_ICON_STOP_DISABLING_SCREEN_SAVER),IMAGE_ICON,16,16,LR_DEFAULTCOLOR|LR_SHARED);
    HICON hIconApp = (HICON)::LoadImage(this->hInstance,MAKEINTRESOURCE(IDI_ICON_APPLICATION),IMAGE_ICON,16,16,LR_DEFAULTCOLOR|LR_SHARED);
    HICON hIconWebSite = (HICON)::LoadImage(this->hInstance,MAKEINTRESOURCE(IDI_ICON_WEB),IMAGE_ICON,16,16,LR_DEFAULTCOLOR|LR_SHARED);

    this->pNotifyIcon = new CNotifyIcon(this->hInstance,hIconApp,TEXT("DisableScreenSaver"),this);

    // create popup menu associated to notify icon
    this->pPopUpMenu = new CPopUpMenu();

    this->MenuIdStartStop = this->pPopUpMenu->Add(STOP_DISABLING,this->hIconStopDisabling);
    this->MenuIdWebSite =  this->pPopUpMenu->Add(TEXT("Home Page"),hIconWebSite);
    this->pPopUpMenu->AddSeparator();
    this->MenuIdExit = this->pPopUpMenu->Add(TEXT("Exit"),IDI_ICON_EXIT,this->hInstance);

    // display notify icon
    this->pNotifyIcon->Show();

    // start disabling
    this->bScreenSaverDisabled = FALSE;
    this->OnStartStop();

    // wait exit message
    this->WaitApplicationEnd();
    
    // destroy objects after application main loop end
    delete pNotifyIcon;
    delete this->pPopUpMenu;

    return TRUE;
}

//void CDisableScreenSaver::OnInit()
//{
//
//}
void CDisableScreenSaver::OnClose()
{
    // if screen saver is disabled, re enable it
    if (this->bScreenSaverDisabled)
        this->OnStartStop();

    ::DestroyWindow(this->hWndDialog);
    this->hWndDialog=NULL;
    ::PostQuitMessage(0);
}

LRESULT CALLBACK CDisableScreenSaver::WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch(uMsg)
    {
    case WM_CREATE:
        {
            CREATESTRUCT* pCreateStruct;
            CDisableScreenSaver* pDialogObject; 
            pCreateStruct=(CREATESTRUCT*)lParam;
            pDialogObject=(CDisableScreenSaver*)pCreateStruct->lpCreateParams;
            SetWindowLongPtr(hWnd,GWLP_USERDATA,(LONG_PTR)pDialogObject);
            pDialogObject->hWndDialog=hWnd;

            // pDialogObject->OnInit();
        }
        break;
    case WM_QUERYENDSESSION:
        return TRUE;
    case WM_QUIT:
    case WM_CLOSE:
        {
            CDisableScreenSaver* pDialogObject; 
            pDialogObject=CDisableScreenSaver::GetAssociatedObject(hWnd);
            if (!pDialogObject)
                break;

            pDialogObject->OnClose();
            SetWindowLongPtr(hWnd,GWLP_USERDATA,(LONG_PTR)NULL);
        }
        break;

    default:
        return ::DefWindowProc(hWnd, uMsg, wParam, lParam);
    }
    return 0;
}

BOOL CDisableScreenSaver::Exit()
{
    if (!this->hWndDialog)
        return FALSE;

    // use post message to allow current message to be fully processed before closing
    ::PostMessage(this->hWndDialog,WM_CLOSE,0,0);

    return TRUE;
}

BOOL CDisableScreenSaver::WaitApplicationEnd()
{
    WNDCLASSEX wc;
    HWND hwnd;
    MSG Msg;

    // if CDisableScreenSaver_CLASS_NAME is not already registered
    if (!GetClassInfoEx(this->hInstance, CDisableScreenSaver_CLASS_NAME , &wc))
    {
        // register it

        wc.cbSize        = sizeof(WNDCLASSEX);
        wc.style         = 0;
        wc.lpfnWndProc   = CDisableScreenSaver::WndProc;
        wc.cbClsExtra    = 0;
        wc.cbWndExtra    = 0;
        wc.hInstance     = this->hInstance;
        wc.hIcon         = NULL;
        wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
        wc.hbrBackground = (HBRUSH)COLOR_WINDOWFRAME;
        wc.lpszMenuName  = NULL;
        wc.lpszClassName = CDisableScreenSaver_CLASS_NAME;
        wc.hIconSm       = NULL;
        if(!RegisterClassEx(&wc))
        {
            ::MessageBox(NULL, _T("Window Registration Failed"), _T("Error"),MB_OK|MB_ICONERROR | MB_SYSTEMMODAL);
            return FALSE;
        }
    }

    // create the dialog window
    hwnd = ::CreateWindowEx(0,
                            CDisableScreenSaver_CLASS_NAME,
                            _T(""),
                            WS_DLGFRAME,
                            0,
                            0, 
                            0, 
                            0,
                            HWND_DESKTOP,
                            NULL,
                            this->hInstance,
                            this);

    if(hwnd == NULL)
    {
        ::MessageBox(NULL, _T("Window Creation Failed"), _T("Error"),MB_OK|MB_ICONERROR | MB_SYSTEMMODAL);
        return FALSE;
    }

    BOOL bRet;
    while( (bRet = ::GetMessage( &Msg, NULL, 0, 0 )) != 0)
    { 
        if (bRet == -1)
        {
            // handle the error and possibly exit
        }
        else
        {
            ::TranslateMessage(&Msg); 
            ::DispatchMessage(&Msg); 
        }
    }
    // return (DWORD)Msg.wParam;
    return TRUE;
}

CDisableScreenSaver* CDisableScreenSaver::GetAssociatedObject(HWND hWndDialog)
{
    return (CDisableScreenSaver*)GetWindowLongPtr(hWndDialog,GWLP_USERDATA);
}
